import pdb
import time
import numpy as np
import copy

from . import gl_
from . import ipge_
from .util_ import print_log, compute_zeta
from .util_ import Param, sparse_matrix

from scipy.linalg import inv, svd
from scipy.sparse import csr_matrix
from scipy.sparse.linalg import svds
from sklearn.neighbors import NearestNeighbors
from scipy.spatial.distance import pdist, squareform

import multiprocess as mp

class LocalViews:
    def __init__(self, exit_at=None, verbose=True, debug=False):
        self.exit_at = exit_at
        self.logs = verbose
        self.debug = debug
        
        
        self.epsilon = None
        self.U = None
        self.local_param_pre = None
        self.local_param_post = None
        
        # For LDLE
        self.GL = None
        self.IPGE = None
        self.gamma = None
        self.phi = None
        
        self.local_start_time = time.time()
        self.global_start_time = time.time()
        
    def log(self, s='', log_time=False):
        if self.logs:
            self.local_start_time = print_log(s, log_time,
                                              self.local_start_time, 
                                              self.global_start_time)
            
    def fit(self, d, X, d_e, neigh_dist, neigh_ind, ddX, local_opts):
        if local_opts['algo'] == 'LDLE':
            self.log('Constructing ' + local_opts['gl_type'] + ' graph Laplacian + its eigendecomposition.')
            GL = gl_.GL()
            GL.fit(neigh_dist, neigh_ind, local_opts)
            self.log('Done.', log_time=True)
            #############################################
            
            # Local views in the ambient space
            epsilon = neigh_dist[:,[local_opts['k']-1]]
            U = sparse_matrix(neigh_ind[:,:local_opts['k']],
                              np.ones((neigh_ind.shape[0],local_opts['k']), dtype=bool))
            #U = U.maximum(U.transpose()) # Idk why I wrote this. This is not needed.
            #############################################

            self.log('Computing Atilde: Inner Prod of Grad of EigFuncs.')
            IPGE = ipge_.IPGE(debug=self.debug)
            if local_opts['Atilde_method'] == 'LLR':
                print('Using LLR')
            elif local_opts['Atilde_method'] == 'LDLE_2':
                print('Using LDLE_2')
            elif local_opts['Atilde_method'] == 'LDLE_3':
                print('Using LDLE_3')
            else: #fem
                IPGE.fem(opts = {'phi': GL.phi, 'neigh_ind': neigh_ind, 'k': local_opts['k'],
                                 'neigh_dist': neigh_dist, 'epsilon': epsilon,
                                 'd': d, 'p': local_opts['p']})
            self.log('Done.', log_time=True)
            
            if self.exit_at == 'post-Atilde':
                return
            #############################################
            
            # Compute gamma
            if local_opts['no_gamma']:
                gamma = np.ones((GL.phi.shape[0], local_opts['N']))
            else:
                gamma = np.sqrt(local_opts['k']/U.dot(GL.phi**2))
            #############################################

            # Compute LDLE: Low Distortion Local Eigenmaps
            self.log('Computing LDLE.')
            local_param_pre = self.compute_LDLE(d, d_e, neigh_ind, neigh_dist, GL.phi, U, IPGE.Atilde, gamma, local_opts)
            self.log('Done.', log_time=True)
            #############################################

            if local_opts['to_postprocess']:
                self.log('Posprocessing LDLE.')
                local_param_post = self.postprocess_LDLE(d_e, neigh_ind, neigh_dist, local_param_pre, U, local_opts)
                self.log('Done.', log_time=True)
            else:
                local_param_post = local_param_pre
            #############################################
            # set b
            local_param_post.b = np.ones(X.shape[0])
            for k in range(X.shape[0]):
                d_e_k = d_e[np.ix_(neigh_ind[k,:],neigh_ind[k,:])].toarray()
                Psi_k = local_param_post.eval_({'view_index': k, 'data_mask': neigh_ind[k,:]})
                local_param_post.b[k] = np.median(squareform(d_e_k))/np.median(pdist(Psi_k))
            
            #############################################
            
            if ddX is not None:
                self.log('Halving objects.')
                n = X.shape[0]
                U = U[:n,:n]
                GL.phi = GL.phi[:n,:]
                IPGE.Atilde = IPGE.Atilde[:n,:,:]
                gamma = gamma[:n,:]
                
                local_param_pre.Psi_i = local_param_pre.Psi_i[:n,:]
                local_param_pre.Psi_gamma = local_param_pre.Psi_gamma[:n,:]
                local_param_pre.zeta = local_param_pre.zeta[:n]
                
                local_param_post.Psi_i = local_param_post.Psi_i[:n,:]
                local_param_post.Psi_gamma = local_param_post.Psi_gamma[:n,:]
                local_param_post.zeta = local_param_post.zeta[:n]
                local_param_post.b = local_param_post.b[:n]
                
                d_e_ = d_e[:n,:n]
                
                for k in range(n):
                    U_k = U[k,:]
                    local_param_pre.zeta[k] = compute_zeta(d_e_[np.ix_(U_k,U_k)],
                                                           local_param_pre.eval_({'view_index': k,
                                                                                  'data_mask': U_k}))
                    local_param_post.zeta[k] = compute_zeta(d_e_[np.ix_(U_k,U_k)],
                                                            local_param_post.eval_({'view_index': k,
                                                                                    'data_mask': U_k}))
                self.log('Done.', log_time=True)
            #############################################
        else:
            self.log('Constructing local views using LTSA.')
            # Local views in the ambient space
            epsilon = neigh_dist[:,[local_opts['k']-1]]
            U = sparse_matrix(neigh_ind[:,:local_opts['k']-1],
                              np.ones((neigh_ind.shape[0],local_opts['k']-1), dtype=bool))
            local_param_pre = None
            local_param_post = self.compute_LTSAP(d, X, d_e, neigh_ind)
            local_param_post.b = np.ones(X.shape[0])
            self.log('Done.', log_time=True)
            
        print('Max local distortion =', np.max(local_param_post.zeta))
        if self.debug:
            if local_opts['algo'] == 'LDLE':
                self.GL = GL
                self.IPGE = IPGE
                self.gamma = gamma
            
            self.epsilon = epsilon
            self.local_param_pre = local_param_pre
        
        if local_opts['algo'] == 'LDLE':
            self.phi = GL.phi
            
        self.U = U
        self.local_param_post = local_param_post
    
    def compute_LDLE(self, d, d_e, neigh_ind, neigh_dist, phi, U, Atilde, gamma, local_opts, print_prop = 0.25):
        n, N = phi.shape
        N = phi.shape[1]
        tau = local_opts['tau']
        delta = local_opts['delta']
        
        print_freq = np.int(n*print_prop)
        
        local_param = Param('LDLE')
        local_param.phi = phi
        local_param.Psi_gamma = np.zeros((n,d))
        local_param.Psi_i = np.zeros((n,d),dtype='int')
        local_param.zeta = np.zeros(n)
        n_proc = local_opts['n_proc']
        
        def target_proc(p_num, chunk_sz, q_):
            start_ind = p_num*chunk_sz
            if p_num == (n_proc-1):
                end_ind = n
            else:
                end_ind = (p_num+1)*chunk_sz

            for k in range(start_ind, end_ind):
                # to store i_1, ..., i_d
                i = np.zeros(d, dtype='int')

                # Grab the precomputed U_k, Atilde_{kij}, gamma_{ki}
                U_k = U[k,:]
                Atilde_k = Atilde[k,:,:]
                gamma_k = gamma[k,:]

                # Compute theta_1
                Atikde_kii = Atilde_k.diagonal()
                theta_1 = np.percentile(Atikde_kii, tau)

                # Compute Stilde_k
                Stilde_k = Atikde_kii >= theta_1

                # Compute i_1
                r_1 = np.argmax(Stilde_k) # argmax finds first index with max value
                temp = gamma_k * np.abs(Atilde_k[:,r_1])
                alpha_1 = np.max(temp * Stilde_k)
                i[0] = np.argmax((temp >= delta*alpha_1) & (Stilde_k))

                for s in range(1,d):
                    i_prev = i[0:s]
                    # compute temp variable to help compute Hs_{kij} below
                    temp = inv(Atilde_k[np.ix_(i_prev,i_prev)])

                    # Compute theta_s
                    Hs_kii = Atikde_kii - np.sum(Atilde_k[:,i_prev] * np.dot(temp, Atilde_k[i_prev,:]).T, 1)
                    temp_ = Hs_kii[Stilde_k]
                    theta_s = np.percentile(temp_, tau)

                    #theta_s=np.max([theta_s,np.min([np.max(temp_),1e-4])])

                    # Compute i_s
                    r_s = np.argmax((Hs_kii>=theta_s) & Stilde_k)
                    Hs_kir_s = Atilde_k[:,[r_s]] - np.dot(Atilde_k[:,i_prev],
                                                          np.dot(temp, Atilde_k[i_prev,r_s][:,np.newaxis]))
                    temp = gamma_k * np.abs(Hs_kir_s.flatten())
                    alpha_s = np.max(temp * Stilde_k)
                    i[s]=np.argmax((temp >= delta*alpha_s) & Stilde_k);

                # Compute Psi_k
                local_param.Psi_gamma[k,:] = gamma_k[i]
                local_param.Psi_i[k,:] = i

                # Compute zeta_{kk}
                d_e_k = d_e[np.ix_(neigh_ind[k,:],neigh_ind[k,:])].toarray()
                local_param.zeta[k] = compute_zeta(d_e_k,
                                                   local_param.eval_({'view_index': k,
                                                                      'data_mask': neigh_ind[k,:]}))

            q_.put((start_ind, end_ind,
                    local_param.zeta[start_ind:end_ind],
                    local_param.Psi_gamma[start_ind:end_ind,:],
                    local_param.Psi_i[start_ind:end_ind,:]))
        
        q_ = mp.Queue()
        chunk_sz = int(n/n_proc)
        proc = []
        for p_num in range(n_proc):
            proc.append(mp.Process(target=target_proc,
                                   args=(p_num,chunk_sz,q_),
                                   daemon=True))
            proc[-1].start()

        for p_num in range(n_proc):
            start_ind, end_ind, zeta_, Psi_gamma_, Psi_i_ = q_.get()
            local_param.zeta[start_ind:end_ind] = zeta_
            local_param.Psi_gamma[start_ind:end_ind,:] = Psi_gamma_
            local_param.Psi_i[start_ind:end_ind,:] = Psi_i_

        q_.close()
        
        for p_num in range(n_proc):
            proc[p_num].join()
            
        print('local_param: all %d points processed...' % n)
        print("max distortion is %f" % (np.max(local_param.zeta)))
        return local_param
    
    def compute_LTSAP(self, d, X, d_e, neigh_ind, print_prop = 0.25):
        n = neigh_ind.shape[0]
        p = X.shape[1]
        print_freq = int(print_prop * n)
        
        local_param = Param('LTSA')
        local_param.X = X
        local_param.Psi = np.zeros((n,p,d))
        local_param.mu = np.zeros((n,p))
        local_param.zeta = np.zeros(n)

        # iterate over points in the data
        for k in range(n):
            if print_freq and np.mod(k, print_freq)==0:
                print('local_param: %d points processed...' % k)
            
            neigh_ind_k = neigh_ind[k,:]
            # LTSA
            X_k = X[neigh_ind_k,:]
            xbar_k = np.mean(X_k,axis=0)[np.newaxis,:]
            X_k = X_k - xbar_k
            X_k = X_k.T
            if p == d:
                Q_k,Sigma_k,_ = svd(X_k)
            else:
                Q_k,Sigma_k,_ = svds(X_k, d, which='LM')
                
            local_param.Psi[k,:,:] = Q_k[:,:d]
            local_param.mu[k,:] = xbar_k
            
            # Compute zeta_{kk}
            d_e_k = d_e[np.ix_(neigh_ind_k, neigh_ind_k)].toarray()
            local_param.zeta[k] = compute_zeta(d_e_k,
                                               local_param.eval_({'view_index': k,
                                                                  'data_mask': neigh_ind_k}))
            
        print('local_param: all %d points processed...' % n)
        return local_param
    
    def postprocess_LDLE(self, d_e, neigh_ind, neigh_dist, local_param_pre, U, local_opts):
        # initializations
        n = neigh_ind.shape[0]
        local_param = copy.deepcopy(local_param_pre)
        
        N_replaced = n
        itr = 1
        # Extra variable to speed up
        param_changed_old = None
        n_proc = local_opts['n_proc']
        
        while N_replaced:
            new_param_of = np.arange(n, dtype='int')
            param_changed_new = np.zeros(n, dtype=bool)
            if N_replaced > local_opts['pp_n_thresh']: # use multiple processors
                zeta = local_param.zeta
                if param_changed_old is not None:
                    param_changed_old = set(np.where(param_changed_old)[0])
                    
                def target_proc(p_num, chunk_sz, q_):
                    start_ind = p_num*chunk_sz
                    if p_num == (n_proc-1):
                        end_ind = n
                    else:
                        end_ind = (p_num+1)*chunk_sz

                    for k in range(start_ind, end_ind):
                        U_k = U[k,:].nonzero()[1].tolist()
                        if param_changed_old is None:
                            cand_k = U_k
                        else:
                            cand_k = param_changed_old.intersection(U_k)
                        neigh_ind_k = neigh_ind[k,:]
                        d_e_k = d_e[np.ix_(neigh_ind_k,neigh_ind_k)].toarray()
                        for kp in cand_k:
                            Psi_kp_on_U_k = local_param.eval_({'view_index': kp,
                                                               'data_mask': neigh_ind_k})
                            zeta_kkp = compute_zeta(d_e_k, Psi_kp_on_U_k)

                            # if zeta_{kk'} < zeta_{kk}
                            if zeta_kkp < zeta[k]:
                                zeta[k] = zeta_kkp
                                new_param_of[k] = kp
                                param_changed_new[k] = 1

                    q_.put((start_ind, end_ind, zeta[start_ind:end_ind],
                            new_param_of[start_ind:end_ind],
                            param_changed_new[start_ind:end_ind]))
                
                q_ = mp.Queue()
                chunk_sz = int(n/n_proc)
                proc = []
                for p_num in range(n_proc):
                    proc.append(mp.Process(target=target_proc,
                                           args=(p_num,chunk_sz,q_),
                                           daemon=True))
                    proc[-1].start()

                for p_num in range(n_proc):
                    start_ind, end_ind, zeta_, new_param_of_, param_changed_new_ = q_.get()
                    zeta[start_ind:end_ind] = zeta_
                    param_changed_new[start_ind:end_ind] = param_changed_new_
                    new_param_of[start_ind:end_ind] = new_param_of_
                
                q_.close()
                
                for p_num in range(n_proc):
                    proc[p_num].join()
                    
                local_param.zeta = zeta
            else: # do sequentially (single processor)
                # Iterate over all local parameterizations
                # To speed up the process, only consider those neighbors
                # for which the parameterization changed in the prev step
                if param_changed_old is None:
                    cand = U.tocoo()
                else:
                    cand = U.multiply(param_changed_old[None,:])
                    cand.eliminate_zeros()

                cand_i = cand.row
                cand_j = cand.col
                n_ind = cand_i.shape[0]
                ind = 0
                for k in range(n):
                    cand_k = []
                    while (ind < n_ind) and (k == cand_i[ind]):
                        cand_k.append(cand_j[ind])
                        ind += 1

                    neigh_ind_k = neigh_ind[k,:]
                    d_e_k = d_e[np.ix_(neigh_ind_k,neigh_ind_k)].toarray()
                    for kp in cand_k:
                        Psi_kp_on_U_k = local_param.eval_({'view_index': kp,
                                                           'data_mask': neigh_ind_k})
                        zeta_kkp = compute_zeta(d_e_k, Psi_kp_on_U_k)

                        # if zeta_{kk'} < zeta_{kk}
                        if zeta_kkp < local_param.zeta[k]:
                            local_param.zeta[k] = zeta_kkp
                            new_param_of[k] = kp
                            param_changed_new[k] = 1
            
            local_param.Psi_i = local_param.Psi_i[new_param_of,:]
            local_param.Psi_gamma = local_param.Psi_gamma[new_param_of,:]
            param_changed_old = param_changed_new.copy()
            N_replaced = np.sum(param_changed_new)
            
            print("Iter %d, Param replaced: %d, max distortion: %f" % (itr, N_replaced, np.max(local_param.zeta)))
            itr = itr + 1
        
        return local_param


import pdb
import numpy as np
import time

from .util_ import procrustes

import scipy
from sklearn.neighbors import NearestNeighbors
from scipy.spatial.distance import pdist, squareform

import multiprocess as mp

# Computes Z_s for the case when to_tear is True.
# Input Z_s is the Z_s for the case when to_tear is False.
# Output Z_s is a subset of input Z_s.
def compute_Z_s_to_tear(y, s, Z_s, C, c, k):
    n_Z_s = Z_s.shape[0]
    # C_s_U_C_Z_s = (self.C[s,:]) | np.isin(self.c, Z_s)
    C_s_U_C_Z_s = (C[s,:]) | np.any(C[Z_s,:], 0)
    c = c[C_s_U_C_Z_s]
    n_ = c.shape[0]

    d_e_ = squareform(pdist(y[C_s_U_C_Z_s,:]))

    k_ = min(k,d_e_.shape[0]-1)
    neigh = NearestNeighbors(n_neighbors=k_,
                             metric='precomputed',
                             algorithm='brute')
    neigh.fit(d_e_)
    neigh_dist, _ = neigh.kneighbors()

    epsilon = neigh_dist[:,[k_-1]]
    Ug = d_e_ < (epsilon + 1e-12)

    Utildeg = np.zeros((n_Z_s+1,n_))
    for m in range(n_Z_s):
        Utildeg[m,:] = np.any(Ug[c==Z_s[m],:], 0)

    Utildeg[n_Z_s,:] = np.any(Ug[c==s,:], 0)

    # |Utildeg_{mm'}|
    n_Utildeg_Utildeg = np.matmul(Utildeg, Utildeg.T)
    np.fill_diagonal(n_Utildeg_Utildeg, 0)

    return Z_s[n_Utildeg_Utildeg[-1,:-1]>0]


def sequential_init(seq, rho, y, is_visited_view, d, Utilde, n_Utilde_Utilde,
                    C, c, intermed_param, global_opts, print_freq=1000,
                    ret_contrib_of_views=False):   
    n = Utilde.shape[1]
    if ret_contrib_of_views:
        # Initialization contribution of view i
        # as the points in cluster i
        contrib_of_view = C.copy()
    # Traverse views from 2nd view
    for m in range(1,seq.shape[0]):
        if print_freq and np.mod(m, print_freq)==0:
            print('Initial alignment of %d views completed' % m, flush=True)
        s = seq[m]
        # pth view is the parent of sth view
        p = rho[s]
        Utilde_s = Utilde[s,:]

        
        # If to tear apart closed manifolds
        if global_opts['to_tear']:
            if global_opts['init_algo']['align_w_parent_only']:
                Z_s = [p]
            else:
                # Compute T_s and v_s by aligning
                # the embedding of the overlap Utilde_{sp}
                # due to sth view with that of the pth view
                Utilde_s_p = Utilde_s*Utilde[p,:]
                V_s_p = intermed_param.eval_({'view_index': s, 'data_mask': Utilde_s_p})
                V_p_s = intermed_param.eval_({'view_index': p, 'data_mask': Utilde_s_p})
                intermed_param.T[s,:,:], intermed_param.v[s,:] = procrustes(V_s_p, V_p_s)

                # Compute temporary global embedding of point in sth cluster
                C_s = C[s,:]
                y[C_s,:] = intermed_param.eval_({'view_index': s, 'data_mask': C_s})

                # Find more views to align sth view with
                Z_s = is_visited_view & (n_Utilde_Utilde[s,:]>0)
                Z_s_all = np.where(Z_s)[0]
                Z_s = compute_Z_s_to_tear(y, s, Z_s_all, C, c, global_opts['k'])
        # otherwise
        else:
            # Align sth view with all the views which have
            # an overlap with sth view in the ambient space
            Z_s = is_visited_view & (n_Utilde_Utilde[s,:]>0)
            Z_s = np.where(Z_s)[0]

            Z_s = Z_s.tolist()
            # If for some reason Z_s is empty
            if len(Z_s)==0:
                Z_s = [p]
                
        # Compute centroid mu_s
        # n_Utilde_s_Z_s[k] = #views in Z_s which contain
        # kth point if kth point is in the sth view, else zero
        n_Utilde_s_Z_s = np.zeros(n, dtype=int)
        mu_s = np.zeros((n,d))
        for mp in Z_s:
            Utilde_s_mp = Utilde_s & Utilde[mp,:]
            if ret_contrib_of_views:
                #overlaps[s,:] = overlaps[s,:] | Utilde_s_mp
                #overlaps[mp,:] = overlaps[mp,:] | (Utilde[mp,:] & C[s,:])
                contrib_of_view[s,:] = contrib_of_view[s,:] | (Utilde_s & C[mp,:])
                
            n_Utilde_s_Z_s[Utilde_s_mp] += 1
            mu_s[Utilde_s_mp,:] += intermed_param.eval_({'view_index': mp, 'data_mask': Utilde_s_mp})

        # Compute T_s and v_s by aligning the embedding of the overlap
        # between sth view and the views in Z_s, with the centroid mu_s
        temp = n_Utilde_s_Z_s > 0
        mu_s = mu_s[temp,:] / n_Utilde_s_Z_s[temp,np.newaxis]
        V_s_Z_s = intermed_param.eval_({'view_index': s, 'data_mask': temp})

        T_s, v_s = procrustes(V_s_Z_s, mu_s)

        # Update T_s, v_
        intermed_param.T[s,:,:] = np.matmul(intermed_param.T[s,:,:], T_s)
        intermed_param.v[s,:] = np.matmul(intermed_param.v[s,:][np.newaxis,:], T_s) + v_s

        # Mark sth view as visited
        is_visited_view[s] = True

        # Compute global embedding of point in sth cluster
        C_s = C[s,:]
        y[C_s,:] = intermed_param.eval_({'view_index': s, 'data_mask': C_s})
    
    if ret_contrib_of_views:
        return y, is_visited_view, contrib_of_view
    else:
        return y, is_visited_view


# Ngoc-Diep Ho, Paul Van Dooren, On the pseudo-inverse of the Laplacian of a bipartite graph
def compute_Lpinv(Utilde):
    B_ = Utilde.T.astype('int')
    D_1 = np.sum(B_, axis=1, keepdims=True)
    D_2 = np.sum(B_, axis=0, keepdims=True)
    D_1_inv_sqrt = np.sqrt(1/D_1)
    D_2_inv_sqrt = np.sqrt(1/D_2)
    B_tilde = D_1_inv_sqrt*B_*D_2_inv_sqrt
    U12,SS,VT = scipy.linalg.svd(B_tilde, full_matrices=False)
    V = VT.T
    mask = np.abs(SS-1)<1e-6
    m_1 = np.sum(mask)
    Sigma = np.expand_dims(SS[m_1:], 1)
    Sigma_1 = 1/(1-Sigma**2)
    Sigma_2 = Sigma*Sigma_1
    U1 = U12[:,:m_1]
    U2 = U12[:,m_1:]
    V1 = V[:,:m_1]
    V2 = V[:,m_1:]
    
    def target_proc(i):
        if i == 0:
            tmp = -0.75*np.matmul(U1,U1.T) + np.matmul(U2, ((Sigma_1-1))*(U2.T)) + np.eye(U12.shape[0])
            tmp = tmp*D_1_inv_sqrt*(D_1_inv_sqrt.T)
        elif i==1:
            tmp = -0.25*np.matmul(U1,V1.T) + np.matmul(U2, Sigma_2*(V2.T))
            tmp = tmp*D_1_inv_sqrt*D_2_inv_sqrt
        elif i==2:
            tmp = 0.25*np.matmul(V1,V1.T) + np.matmul(V2, Sigma_1*(V2.T))
            tmp = tmp*(D_2_inv_sqrt.T)*D_2_inv_sqrt
        return tmp, np.sum(tmp, axis=1, keepdims=True), np.sum(tmp, axis=0, keepdims=True)
    
    L_tilde = []
    L_tilde_sum_1 = []
    L_tilde_sum_0 = []
    ################################################
    # Parallel
#     with mp.Pool(3) as p:
#         tmp = p.map(target_proc, [0,1,2])
#         for elem in tmp:
#             L_tilde.append(elem[0])
#             L_tilde_sum_1.append(elem[1])
#             L_tilde_sum_0.append(elem[2])
        
    # Sequential
    for i in range(3):
        elem = target_proc(i)
        L_tilde.append(elem[0])
        L_tilde_sum_1.append(elem[1])
        L_tilde_sum_0.append(elem[2])
    ################################################
    
    L_pinv_1 = np.concatenate((L_tilde[0], L_tilde[1]), axis=1)
    L_pinv_2 = np.concatenate((L_tilde[1].T, L_tilde[2]), axis=1)
    Lpinv = np.concatenate([L_pinv_1, L_pinv_2], axis=0)

    n1 = Lpinv.shape[1]
    v1 = np.concatenate((L_tilde_sum_1[0]+L_tilde_sum_1[1],
                         L_tilde_sum_0[1].T+L_tilde_sum_1[2]), axis=0)
    n0 = Lpinv.shape[0]
    v0 = np.concatenate((L_tilde_sum_0[0]+L_tilde_sum_1[1].T,
                         L_tilde_sum_0[1]+L_tilde_sum_0[2]), axis=1)
    Lpinv = Lpinv - v1/n1 - v0.T/n0
    return Lpinv
    

def compute_CC(D, B, Lpinv):
    CC = D - np.matmul(B, np.matmul(Lpinv, B.T))
    CC = 0.5*(CC + CC.T)
    return CC

    
def build_ortho_optim(d, Utilde, intermed_param, tol = 1e-6, Lpinv=None):
    M,n = Utilde.shape
    B = np.zeros((M*d,n+M))
    D = np.zeros((M*d,M*d))
    for i in range(M):
        X_ = intermed_param.eval_({'view_index': i,
                                   'data_mask': Utilde[i,:]})
        D[d*i:d*(i+1),d*i:d*(i+1)] = np.matmul(X_.T,X_)
        B[d*i:d*(i+1),n+i] = -np.sum(X_.T, axis=1)
        B[d*i:d*(i+1),np.where(Utilde[i,:])[0]] = X_.T

    
    if Lpinv is None:
        print('Computing Pseudoinverse of a matrix of L of size', n+M, flush=True)
        Lpinv = compute_Lpinv(Utilde)
    
    CC = compute_CC(D, B, Lpinv)
    return CC, Lpinv, B
    

def compute_alignment_err(d, Utilde, intermed_param, tol = 1e-6, CC=None, Lpinv=None, B=None):
    if (CC is None) or (Lpinv is None) or (B is None):
        CC, Lpinv, B = build_ortho_optim(d, Utilde, intermed_param, tol = 1e-6)
    else:
        CC, Lpinv, B = build_ortho_optim(d, Utilde, intermed_param, tol = 1e-6, Lpinv=Lpinv)
    err = 0
    M,n = Utilde.shape
    CC_mask = np.tile(np.eye(d, dtype=bool), (M,M))
    err = np.sum(CC[CC_mask])
    return err, [CC, Lpinv, B]
    
def spectral_init(y, is_visited_view, d, Utilde,
                  C, intermed_param, global_opts, print_freq=1000,
                  tol = 1e-6):
    CC, Lpinv, B = build_ortho_optim(d, Utilde, intermed_param, tol = 1e-6)
    M,n = Utilde.shape
    print('Computing eigh(C,k=d)', flush=True)
    np.random.seed(42)
    v0 = np.random.uniform(0,1,CC.shape[0])
    # To find smallest eigenvalues, using shift-inverted algo with mode=normal and which='LM'
    W_,V_ = scipy.sparse.linalg.eigsh(CC, k=d, v0=v0, sigma=0.0)
    print('Done.', flush=True)
    Wstar = np.sqrt(M)*V_.T
    
    Tstar = np.zeros((d, M*d))
    for i in range(M):
        U_,S_,VT_ = scipy.linalg.svd(Wstar[:,d*i:d*(i+1)])
        temp_ = np.matmul(U_,VT_)
        if (global_opts['init_algo']['name'] != 'spectral') and (np.linalg.det(temp_) < 0):
            VT_[-1,:] *= -1
            Tstar[:,i*d:(i+1)*d] = np.matmul(U_, VT_)
        else:
            Tstar[:,i*d:(i+1)*d] = temp_
    
    Zstar = np.matmul(Tstar, np.matmul(B, Lpinv))
    
    for s in range(M):
        T_s = Tstar[:,s*d:(s+1)*d].T
        v_s = Zstar[:,n+s]
        intermed_param.T[s,:,:] = np.matmul(intermed_param.T[s,:,:], T_s)
        intermed_param.v[s,:] = np.matmul(intermed_param.v[s,:][np.newaxis,:], T_s) + v_s
        C_s = C[s,:]
        y[C_s,:] = intermed_param.eval_({'view_index': s, 'data_mask': C_s})
        is_visited_view[s] = 1
    
    return y, Zstar[:,:n].T, is_visited_view

def sequential_final(y, d, Utilde, C, intermed_param, n_Utilde_Utilde, n_Utildeg_Utildeg,
                     first_intermed_view_in_cluster,
                     parents_of_intermed_views_in_cluster, cluster_of_intermed_view,
                     global_opts):
    M,n = Utilde.shape
    # Traverse over intermediate views in a random order
    seq = np.random.permutation(M)

    # For a given seq, refine the global embedding
    for it1 in range(global_opts['refine_algo']['max_internal_iter']):
        for s in seq.tolist():
            # Never refine s_0th intermediate view
            if s in first_intermed_view_in_cluster:
                C_s = C[s,:]
                y[C_s,:] = intermed_param.eval_({'view_index': s, 'data_mask': C_s})
                continue

            Utilde_s = Utilde[s,:]

            # If to tear apart closed manifolds
            if global_opts['to_tear']:
                # Find more views to align sth view with
                Z_s = (n_Utilde_Utilde[s,:] > 0) & (n_Utildeg_Utildeg[s,:] > 0)
            # otherwise
            else:
                # Align sth view with all the views which have
                # an overlap with sth view in the ambient space
                Z_s = n_Utilde_Utilde[s,:] > 0

            Z_s = np.where(Z_s)[0].tolist()

            if len(Z_s) == 0:
                Z_s = parents_of_intermed_views_in_cluster[cluster_of_intermed_view[s]][s]
                Z_s = [Z_s]

            # Compute centroid mu_s
            # n_Utilde_s_Z_s[k] = #views in Z_s which contain
            # kth point if kth point is in the sth view, else zero
            n_Utilde_s_Z_s = np.zeros(n, dtype=int)
            mu_s = np.zeros((n,d))
            for mp in Z_s:
                Utilde_s_mp = Utilde_s & Utilde[mp,:]
                n_Utilde_s_Z_s[Utilde_s_mp] += 1
                mu_s[Utilde_s_mp,:] += intermed_param.eval_({'view_index': mp, 'data_mask': Utilde_s_mp})

            temp = n_Utilde_s_Z_s > 0
            mu_s = mu_s[temp,:] / n_Utilde_s_Z_s[temp,np.newaxis]

            # Compute T_s and v_s by aligning the embedding of the overlap
            # between sth view and the views in Z_s, with the centroid mu_s
            V_s_Z_s = intermed_param.eval_({'view_index': s, 'data_mask': temp})
            
            T_s, v_s = procrustes(V_s_Z_s, mu_s)

            # Update T_s, v_s
            intermed_param.T[s,:,:] = np.matmul(intermed_param.T[s,:,:], T_s)
            intermed_param.v[s,:] = np.matmul(intermed_param.v[s,:][np.newaxis,:], T_s) + v_s

            # Compute global embedding of points in sth cluster
            C_s = C[s,:]
            y[C_s,:] = intermed_param.eval_({'view_index': s, 'data_mask': C_s})
    return y

def retraction_final(y, d, Utilde, C, intermed_param,
                     n_Utilde_Utilde, n_Utildeg_Utildeg,
                     first_intermed_view_in_cluster,
                     parents_of_intermed_views_in_cluster,
                     cluster_of_intermed_view,
                     global_opts, CC=None, Lpinv=None, B=None):
    if (CC is None) or (Lpinv is None) or (B is None):
        CC, Lpinv, B = build_ortho_optim(d, Utilde, intermed_param, tol = 1e-6)
    else:
        CC, Lpinv, B = build_ortho_optim(d, Utilde, intermed_param, tol = 1e-6, Lpinv=Lpinv)
    M,n = Utilde.shape
    n_proc = min(M,global_opts['n_proc'])
    def skew(A):
        return 0.5*(A-A.T)

    def unique_qr(A):
        Q, R = np.linalg.qr(A)
        signs = 2 * (np.diag(R) >= 0) - 1
        Q = Q * signs[np.newaxis, :]
        R = R * signs[:, np.newaxis]
        return Q, R

    def update(O, t):
        O = np.copy(O)
        xi = 2*np.matmul(O, CC)
        I_d = np.eye(d)
        
        ###########################################
        # Parallel Updates
        ###########################################
        def target_proc(p_num, chunk_sz, q_):
            start_ind = p_num*chunk_sz
            if p_num == (n_proc-1):
                end_ind = M
            else:
                end_ind = (p_num+1)*chunk_sz
            for i in range(start_ind, end_ind):
                temp0 = O[:,i*d:(i+1)*d]
                temp1 = skew(np.matmul(xi[:,i*d:(i+1)*d],temp0.T))
                Q_,R_ = unique_qr(temp0 - t*np.matmul(temp1,temp0))
                O[:,i*d:(i+1)*d] = Q_

            q_.put((start_ind, end_ind, O[:,start_ind*d:end_ind*d]))
        
        q_ = mp.Queue()
        proc = []
        chunk_sz = int(M/n_proc)
        for p_num in range(n_proc):
            proc.append(mp.Process(target=target_proc,
                                   args=(p_num,chunk_sz,q_),
                                   daemon=True))
            proc[-1].start()

        for p_num in range(n_proc):
            start_ind, end_ind, O_ = q_.get()
            O[:,start_ind*d:end_ind*d] = O_

        q_.close()
        for p_num in range(n_proc):
            proc[p_num].join()
        ###########################################
        
        # Sequential version of above
        # for i in range(M):
        #     temp0 = O[:,i*d:(i+1)*d]
        #     temp1 = skew(np.matmul(xi[:,i*d:(i+1)*d],temp0.T))
        #     Q_,R_ = unique_qr(temp0 - t*np.matmul(temp1,temp0))
        #     O[:,i*d:(i+1)*d] = Q_
        
        return O

    alpha = global_opts['refine_algo']['alpha']
    max_iter = global_opts['refine_algo']['max_internal_iter']
    Tstar = np.zeros((d,M*d))
    for s in range(M):
        Tstar[:,s*d:(s+1)*d] = np.eye(d)
    
    print('Descent starts', flush=True)
    for _ in range(max_iter):
        Tstar = update(Tstar, alpha)
    
    Zstar = np.matmul(Tstar, np.matmul(B, Lpinv))
    
    for s in range(M):
        T_s = Tstar[:,s*d:(s+1)*d].T
        v_s = Zstar[:,n+s]
        intermed_param.T[s,:,:] = np.matmul(intermed_param.T[s,:,:], T_s)
        intermed_param.v[s,:] = np.matmul(intermed_param.v[s,:][np.newaxis,:], T_s) + v_s
        C_s = C[s,:]
        y[C_s,:] = intermed_param.eval_({'view_index': s, 'data_mask': C_s})

    return y, [CC, Lpinv, B]


import pdb
import numpy as np
import time

from scipy.spatial.distance import pdist, squareform
from sklearn.neighbors import NearestNeighbors, KNeighborsTransformer

from scipy.sparse import csr_matrix

from . import local_views_
from . import intermed_views_
from . import global_views_
from . import visualize_
from .util_ import print_log, sparse_matrix, nearest_neighbors
import multiprocessing
    
def double_manifold_k_nn(X, ddX, k_nn):
    """Doubles the manifold represented by X and computes
    k-nearest neighbors of each point on the double.
    
    Parameters
    ----------
    X : array shape (n_samples, n_features)
        2d np.array of shape (n,p)
        where each row represents a point.
    ddX : array shape (n_samples,)
          A 1d np.array of shape (n,) where ddX[i] = 0 iff the
          i-th point X[i,:] represents a point on the boundary
          of the manifold. Note that n_boundary = np.sum(ddX==0).
    k_nn : int
           Number of nearest neighbors to compute on the double.
    
    Returns
    -------
    neigh_dist : array shape (2*n_samples-n_boundary, k_nn)
                 distance of k-nearest neighbors from each
                 point on the double.
    neigh_ind : array shape (2*n_samples-n_boundary, k_nn)
                indices of k-nearest neighbors from each point
                on the double.
    """
    k_nn_ = 2*k_nn
    neigh = KNeighborsTransformer(n_neighbors=k_nn_, n_jobs=-1) #start with 2 x k_nn
    X_dist_graph = neigh.fit_transform(X)
    
    # Compute mask of points on the boundary
    n = X.shape[0]
    dX = ddX==0
    n_dX = np.sum(dX)
    print('No. of points on the boundary =', n_dX)

    # For points not on the boundary, a duplicate point
    # with index offset-ed by n is created
    ind_mask = np.arange(n)
    ind_mask[~dX] = n+np.arange(n-n_dX)
    
    # eliminate the zeros in the original nearest neighbor graph
    # this basically removes the diagonal entries
    X_dist_graph.eliminate_zeros()
    
    # find row and col indices of non-zero entries
    row_inds = []
    col_inds = []
    for row, col in zip(*X_dist_graph.nonzero()):
        row_inds.append(row)
        col_inds.append(col)
        
    row_inds = np.array(row_inds)
    col_inds = np.array(col_inds)
    dist_data = X_dist_graph.data
    
    # a mask of row_inds. True means
    # the row index corresponds to point on the boundary
    dX_inds_mask = dX[row_inds]
    
    # the row indices of points not on the boundary
    # in the duplicated manifold. col indices are computed
    # using ind_mask.
    row_inds2 = ind_mask[row_inds[~dX_inds_mask]]
    col_inds2 = ind_mask[col_inds[~dX_inds_mask]]
    dist_data2 = dist_data[~dX_inds_mask]
    
    # Initialize the doubled graph
    row_inds = np.concatenate([row_inds, row_inds2, np.arange(2*n-n_dX)])
    col_inds = np.concatenate([col_inds, col_inds2, np.arange(2*n-n_dX)])
    dist_data = np.concatenate([dist_data, dist_data2, np.zeros(2*n-n_dX)])
    
    # free up memory
    del row_inds2, col_inds2, dist_data2, X_dist_graph
    
    X_dist_graph2 = csr_matrix((dist_data, (row_inds, col_inds)), shape=(2*n-n_dX, 2*n-n_dX))
    
    # connect points near the boundary on the original
    # and the duplicated manifold
    dX_inds = np.where(dX)[0]
    is_visited = np.zeros((2*n-n_dX,2*n-n_dX), dtype=bool)
    edges = {}
    for i in range(n_dX):
        k = dX_inds[i]
        col_k = X_dist_graph2.getcol(k)
        nbrs = col_k.indices.tolist()
        dists = col_k.data.tolist()
        n_nbrs = len(nbrs)
        for i_ in range(n_nbrs):
            nbr_i = nbrs[i_]
            for j_ in range(i_+1, n_nbrs):
                nbr_j = nbrs[j_]
                if not is_visited[nbr_i,nbr_j]:
                    edges[(nbr_i,nbr_j)] = dists[i_]+dists[j_]
                    edges[(nbr_j,nbr_i)] = edges[(nbr_i,nbr_j)]
                else:
                    edges[(nbr_i,nbr_j)] = np.min([edges[(nbr_i,nbr_j)], dists[i_]+dists[j_]])
                    edges[(nbr_j,nbr_i)] = edges[(nbr_i,nbr_j)]
                    is_visited[nbr_i,nbr_j] = True
                    is_visited[nbr_j,nbr_i] = True
            
    
    del is_visited
    
    row_inds3 = []
    col_inds3 = []
    dist_data3 = []
    for edge, val in edges.items():
        row_inds3.append(edge[0])
        col_inds3.append(edge[1])
        dist_data3.append(val)
    
    # add self connections with approx zero weight
    for i in range(2*n-n_dX):
        row_inds3.append(i)
        col_inds3.append(i)
        dist_data3.append(1e-12)
        
    # Build dinal graph of doubled manifold
    row_inds = np.concatenate([row_inds, np.array(row_inds3)])
    col_inds = np.concatenate([col_inds, np.array(col_inds3)])
    dist_data = np.concatenate([dist_data, np.array(dist_data3)])
    
    # free up memory
    del edges, row_inds3, col_inds3, dist_data3
    
    X_dist_graph3 = csr_matrix((dist_data, (row_inds, col_inds)), shape=(2*n-n_dX, 2*n-n_dX))
    
    pdb.set_trace()

    neigh_dist, neigh_ind = nearest_neighbors(X_dist_graph3, k_nn=k_nn, metric='precomputed')
    return neigh_dist, neigh_ind
    

def get_default_local_opts(algo='LDLE', k_nn=48, k_tune=6, k=24, gl_type='unnorm',
                           N=100, no_gamma=False, Atilde_method='LDLE_1',
                           p=0.99, tau=50, delta=0.9, to_postprocess= True,
                           pp_n_thresh=32):
    """Sets and returns a dictionary of default_local_opts.
    
    Parameters
    ----------
    algo : str
           The algorithm to use for the construction of
           local views. Options are 'LDLE' and 'LTSA'.
           LTSA uses the hyperparameter k only and is
           not affected by the value of the others.
    k_nn : int
           For k-nearest neighbor graph construction.
    k_tune : int
           Distance to k_tune-th nearest neighbor is
           used as the local scaling factor. This must
           be less than k_nn.
    k : int 
        The size of local view per point.
    gl_type : str
              The type of graph Laplacian to construct.
              Options are 'unnorm' for unnormalized,
              'symnorm' for symmetric normalized,
              'diffusion' for density-corrected normalized.
    N : int
        Number of smallest non-trivial eigenvectors of the
        graph Laplacian to be used for the construction of
        local views.
    no_gamma : bool
               If True, the parameterization of local views
               are not normalized.
    Atilde_method : str
                    Method to use for conmputing Atilde.
                    Options are 'LDLE_1' for finite element method,
                    'LLR' for local linear regression. Currently,
                    only 'LDLE_1' works.
    p : float
        A hyperparameter used in computing Atilde. The value
        must be in (0,1).
    tau : int
        A hyperparameter used in computing parameterizations
        of local views. The value must be in (0,100).
    delta : float
        A hyperparameter used in computing parameterizations
        of local views. The value must be in (0,1).
    to_postprocess : bool
        If True the local parameterizations are postprocessed
        to fix anamolous parameterizations leading to high
        distortion of the local views.
    pp_n_thresh : int
        Threshold to use multiple processors or a single processor
        while postprocessing the local parameterizations. A small
        value such as 32 leads to faster postprocessing.
    """
    return {'k_nn': k_nn, 'k_tune': k_tune, 'k': k,
           'gl_type': gl_type, 'N': N, 'no_gamma': no_gamma,
           'Atilde_method': Atilde_method, 'p': p, 'tau': tau,
           'delta': delta, 'to_postprocess': to_postprocess, 'algo': algo,
           'pp_n_thresh': pp_n_thresh}

def get_default_intermed_opts(eta_min=5, eta_max=100, len_S_thresh=256):
    """Sets and returns a dictionary of default_intermed_opts.
    
    Parameters
    ----------
    eta_min : int
              Minimum allowed size of the clusters underlying
              the intermediate views. The values must be >= 1.
    eta_max : int
              Maximum allowed size of the clusters underlying
              the intermediate views. The value must be > eta_min.
    len_S_thresh : int
                   Threshold on the number of points for which 
                   the costs are to be updated, to invoke
                   multiple processors.
    """
    return {'eta_min': eta_min, 'eta_max': eta_max, 'len_S_thresh': len_S_thresh}
    
def get_default_global_opts(main_algo='LDLE', to_tear=True, nu=3, max_iter=10,
                            vis_before_init=False, compute_error=False,
                            init_algo_name='sequential', init_algo_align_w_parent_only=True,
                            refine_algo_name='retraction',
                            refine_algo_max_internal_iter=100,
                            refine_algo_alpha=0.3):
    """Sets and returns a dictionary of default_global_opts.

    Parameters
    ----------
    main_algo : str
                The algorithm to use for the alignment of intermediate
                views. Options are 'LDLE' and 'LTSA'. If 'LTSA' is
                chosen then none of the following hypermateters are used.
    to_tear : bool
              If True the tearing of the manifold is allowed.
    nu : int
         A hyperparameter used to compute neighboring intermediate
         views in the embedding space.
    max_iter : int
               Number of iterations to refine the global embedding for.
    vis_before_init : bool
                      If True, plots the global embedding before
                      alignment begins. This is same as just plotting
                      all the intermediate views without alignment.
    compute_error : bool
                    If True the alignment error is computed at each 
                    iteration of the refinement, otherwise only at
                    the last iteration.
    init_algo_name : str
                     The algorithm used to compute initial global embedding
                     by aligning the intermediate views.
                     Options are 'sequential' for tree-based-procrustes alignment,
                     'spectral' for spectral alignment (ignores to_tear),
                     'sequential+spectral' for tree-based-procrustes alignment
                     followed by spectral alignment.
    init_algo_align_w_parent_only : bool
                                    If True only the parents of the intermediate
                                    views are used in the tree-based-procrustes
                                    alignment.
    refine_algo_name : str
                       The algorithm used to refine the initial global embedding
                       by refining the alignment between intermediate views.
                       Options are 'sequential' for Generalized Procustes Analysis
                       (GPA) based alignment, 'retraction' for Riemannian optimization
                       based alignment, 'spectral' for spectral alignment.
    refine_algo_max_internal_iter : int
                                    The number of internal iterations used by
                                    the refinement algorithm. This is ignored
                                    by 'spectral' refinement.
    refine_algo_alpha : int
                        The step size used in the Riemannian gradient descent
                        when the refinement algorithm is 'retraction'.
    """
    return {'to_tear': to_tear, 'nu': nu, 'max_iter': max_iter,
               'vis_before_init': vis_before_init,
               'compute_error': compute_error,
               'main_algo': main_algo, # ['LDLE', 'LTSA']
               'init_algo': {
                   'name': init_algo_name, # ['sequential', 'spectral', 'sequential+spectral'], spectral ignores to_tear
                   'align_w_parent_only': init_algo_align_w_parent_only
                },
               'refine_algo': {
                   'name': refine_algo_name, # ['sequential', 'retraction', 'spectral']
                   'max_internal_iter': refine_algo_max_internal_iter, # 10 for sequential and 100 for retraction
                   'alpha': refine_algo_alpha, # step size for retraction
                }
              }
def get_default_vis_opts(save_dir='', cmap_interior='summer', cmap_boundary='jet', c=None):
    return {'save_dir': save_dir,
             'cmap_interior': cmap_interior,
             'cmap_boundary': cmap_boundary,
             'c': c}


class LDLE:
    """Low Dimensional Local Eigenmaps and its variations.
    
    Parameters
    ----------
    d : int
       Intrinsic dimension of the manifold.
    local_opts : dict
                Options for local views construction. The key-value pairs
                provided override the ones in default_local_opts.
    intermed_opts : dict
                    Options for intermediate views construction. The key-value pairs
                    provided override the ones in default_intermed_opts.
    global_opts : dict
                  Options for global views construction. The key-value pairs
                  provided override the ones in default_global_opts.
    vis_opts : dict
               Options for visualization. The key-value pairs
               provided override the ones in default_vis_opts.
    n_proc : int
             The number of processors to use. Defaults to approximately
             3/4th of the available processors. 
    verbose : bool
              print logs if True.
    debug : bool
            saves intermediary objects/data for debugging.
    """
    def __init__(self,
                 d = 2,
                 local_opts = {}, 
                 intermed_opts = {},
                 global_opts = {},
                 vis_opts = {},
                 n_proc = max(1,int(multiprocessing.cpu_count()*0.75)),
                 exit_at = None,
                 verbose = False,
                 debug = False):
        default_local_opts = get_default_local_opts()
        default_intermed_opts = get_default_intermed_opts()
        default_global_opts = get_default_global_opts()
        default_vis_opts = get_default_vis_opts()
        
        self.d = d
        local_opts['n_proc'] = n_proc
        for i in local_opts:
            default_local_opts[i] = local_opts[i]
        self.local_opts = default_local_opts
        self.local_opts['k_nn'] = max(self.local_opts['k_nn'],
                                      self.local_opts['k'])
        #############################################
        intermed_opts['algo'] = self.local_opts['algo']
        intermed_opts['n_proc'] = n_proc
        for i in intermed_opts:
            default_intermed_opts[i] = intermed_opts[i]
        self.intermed_opts = default_intermed_opts
        #############################################
        global_opts['k'] = self.local_opts['k']
        global_opts['n_proc'] = n_proc
        for i in global_opts:
            default_global_opts[i] = global_opts[i]
        self.global_opts = default_global_opts
        #############################################
        for i in vis_opts:
            default_vis_opts[i] = vis_opts[i]
        self.vis_opts = default_vis_opts
        self.vis = visualize_.Visualize(self.vis_opts['save_dir'])
            
        #############################################
        self.exit_at = exit_at
        self.verbose = verbose
        self.debug = debug
        #############################################
        
        # Other useful inits
        self.global_start_time = time.time()
        self.local_start_time = time.time()
        
        # The variables created during the fit
        self.X = None
        self.d_e = None
        self.ddX = None
        self.neigh_dist = None
        self.neigh_ind = None
        self.LocalViews = None
        self.IntermedViews = None
        
    def log(self, s='', log_time=False):
        if self.verbose:
            self.local_start_time = print_log(s, log_time,
                                              self.local_start_time, 
                                              self.global_start_time)
    
    def fit(self, X = None, d_e = None, ddX = None):
        """Run the algorithm. Either X or d_e must be supplied.
        
        Parmeters
        ---------
        X : array shape (n_samples, n_features)
            A 2d array containing data representing a manifold.
        d_e : array shape (n_samples, n_samples)
            A square numpy matrix representing the geodesic distance
            between each pair of points on the manifold.
        ddX : array shape (n_samples)
            Optional. A 1d array representing the distance of the
            points from the boundary OR a 1d array such that 
            ddX[i] = 0 if the point i i.e. X[i,:] is on the boundary.
            
        Returns
        -------
        y : array shape (n_samples, d)
            The embedding of the data in lower dimension.
        """
        assert X is not None or d_e is not None, "Either X or d_e should be provided."
        
        if d_e is None:
            if ddX is None or self.local_opts['algo'] == 'LTSA':
                neigh_dist, neigh_ind = nearest_neighbors(X,
                                                          k_nn=self.local_opts['k_nn'],
                                                          metric='euclidean')
            else:
                self.log('Doubling manifold.')
                neigh_dist, neigh_ind = double_manifold_k_nn(X, ddX, self.local_opts['k_nn'])
                self.log('Done.', log_time=True)
        
        # Construct a sparse d_e matrix based on neigh_ind and neigh_dist
        d_e = sparse_matrix(neigh_ind, neigh_dist)
        d_e = d_e.maximum(d_e.transpose())
        
        # Construct low dimensional local views
        LocalViews = local_views_.LocalViews(self.exit_at, self.verbose, self.debug)
        LocalViews.fit(self.d, X, d_e, neigh_dist, neigh_ind, ddX, self.local_opts)
        
        # Halving distance matrix
        if ddX is not None:
            n = ddX.shape[0]
            d_e = d_e[:n,:n]
        
        # Construct intermediate views
        IntermedViews = intermed_views_.IntermedViews(self.exit_at, self.verbose, self.debug)
        IntermedViews.fit(self.d, d_e, neigh_ind[:,:self.local_opts['k']], LocalViews.U,
                          LocalViews.local_param_post, self.intermed_opts)
        
        # Construct Global views
        GlobalViews = global_views_.GlobalViews(self.exit_at, self.verbose, self.debug)
        GlobalViews.fit(self.d, IntermedViews.Utilde, IntermedViews.C, IntermedViews.c,
                        IntermedViews.n_C, IntermedViews.intermed_param, self.global_opts,
                        self.vis, self.vis_opts)
        
        self.LocalViews = LocalViews
        self.IntermedViews = IntermedViews
        self.GlobalViews = GlobalViews
        
        if self.debug:
            self.d_e = d_e
            self.neigh_ind = neigh_ind
            self.neigh_dist = neigh_dist
        
        return GlobalViews.y_final